@@ -117,6 +117,11 @@ static int _pam_parse_conf_file(pam_handle_t *pamh, FILE *f
 		module_type = PAM_T_ACCT;
 	    } else if (!strcasecmp("password", tok)) {
 		module_type = PAM_T_PASS;
+	    } else if (!strcasecmp("@include", tok)) {
+	    	pam_include = 1;
+	    	substack = 0;
+	    	module_type = requested_module_type;
+	    	goto parsing_done;
 	    } else {
 		/* Illegal module type */
 		D(("_pam_init_handlers: bad module type: %s", tok));
@@ -186,6 +191,7 @@ static int _pam_parse_conf_file(pam_handle_t *pamh, FILE *f
 		_pam_set_default_control(actions, _PAM_ACTION_BAD);
 	    }
 
+parsing_done:
 	    tok = _pam_StrTok(NULL, " \n\t", &nexttok);
 	    if (pam_include) {
 	    	if (substack) {
@@ -511,6 +517,7 @@ int _pam_init_handlers(pam_handle_t *pamh)
 static int _pam_assemble_line(FILE *f, char *buffer, int buf_len)
 {
     char *p = buffer;
+    char *endp = buffer + buf_len;
     char *s, *os;
     int used = 0;
 
@@ -518,12 +525,12 @@ static int _pam_assemble_line(FILE *f, char *buffer, int buf_len)
 
     D(("called."));
     for (;;) {
-	if (used >= buf_len) {
+	if (p >= endp) {
 	    /* Overflow */
 	    D(("_pam_assemble_line: overflow"));
 	    return -1;
 	}
-	if (fgets(p, buf_len - used, f) == NULL) {
+	if (fgets(p, endp - p, f) == NULL) {
 	    if (used) {
 		/* Incomplete read */
 		return -1;
