@@ -37,7 +37,9 @@
 #include "config.h"
 
 #include <stdio.h>
-#ifdef HAVE_CRYPT_H
+#ifdef HAVE_LIBXCRYPT
+# include <xcrypt.h>
+#elif defined(HAVE_CRYPT_H)
 # include <crypt.h>
 #endif
 #include <unistd.h>
@@ -49,6 +51,7 @@
 #include <sys/stat.h>
 #include <ctype.h>
 #include <limits.h>
+#include <errno.h>
 
 #ifdef HAVE_CRACK_H
 #include <crack.h>
@@ -57,8 +60,12 @@ extern char *FascistCheck(char *pw, const char *dictpath);
 #endif
 
 #ifndef CRACKLIB_DICTS
+#ifdef HAVE_CRACKLIB_DICTPATH
+#define CRACKLIB_DICTS CRACKLIB_DICTPATH
+#else
 #define CRACKLIB_DICTS NULL
 #endif
+#endif
 
 /* For Translators: "%s%s" could be replaced with "<service> " or "". */
 #define PROMPT1 _("New %s%spassword: ")
@@ -98,6 +105,7 @@ struct cracklib_options {
 	int oth_credit;
         int min_class;
 	int use_authtok;
+	int cluster;
 	char prompt_type[BUFSIZ];
         const char *cracklib_dictpath;
 };
@@ -112,6 +120,7 @@ struct cracklib_options {
 #define CO_LOW_CREDIT   1
 #define CO_OTH_CREDIT   1
 #define CO_USE_AUTHTOK  0
+#define CO_CLUSTER	0
 
 static int
 _pam_parse (pam_handle_t *pamh, struct cracklib_options *opt,
@@ -169,6 +178,8 @@ _pam_parse (pam_handle_t *pamh, struct cracklib_options *opt,
                      opt->min_class = 4 ;
 	 } else if (!strncmp(*argv,"use_authtok",11)) {
 		 opt->use_authtok = 1;
+	 } else if (!strncmp(*argv,"cluster",7)) {
+		 opt->cluster = 1;
 	 } else if (!strncmp(*argv,"dictpath=",9)) {
 	     opt->cracklib_dictpath = *argv+9;
 	     if (!*(opt->cracklib_dictpath)) {
@@ -473,15 +484,32 @@ static const char *password_check(struct cracklib_options *opt,
 
 
 #define OLD_PASSWORDS_FILE	"/etc/security/opasswd"
+#define CLUSTER_ROOTDIR		"/etc/cluster_rootdir"
 
-static const char * check_old_password(const char *forwho, const char *newpass)
+static const char * check_old_password(pam_handle_t *pamh,
+				       struct cracklib_options *opt,
+				       const char *forwho, const char *newpass)
 {
 	static char buf[16384];
+	char opasswd[1024+21] = "";
+	int pos;
 	char *s_luser, *s_uid, *s_npas, *s_pas;
 	const char *msg = NULL;
 	FILE *opwfile;
 
-	opwfile = fopen(OLD_PASSWORDS_FILE, "r");
+	if (opt->cluster) {
+		pos = readlink(CLUSTER_ROOTDIR, opasswd, 1024 - 1);
+		if (pos > 0)
+			opasswd[pos] = '\0';
+		else {
+			pam_syslog(pamh, LOG_ERR, "Invalid " CLUSTER_ROOTDIR ": %s",
+				   strerror(errno));
+			return NULL;
+		}
+	}
+	strcat(opasswd, OLD_PASSWORDS_FILE);
+	
+	opwfile = fopen(opasswd, "r");
 	if (opwfile == NULL)
 		return NULL;
 
@@ -539,7 +567,7 @@ static int _pam_unix_approve_pass(pam_handle_t *pamh,
 		pam_syslog(pamh,LOG_ERR,"Can not get username");
 	    return PAM_AUTHTOK_ERR;
 	}
-	msg = check_old_password(user, pass_new);
+	msg = check_old_password(pamh, opt, user, pass_new);
     }
 
     if (msg) {
@@ -574,6 +602,7 @@ PAM_EXTERN int pam_sm_chauthtok(pam_handle_t *pamh, int flags,
     options.low_credit = CO_LOW_CREDIT;
     options.oth_credit = CO_OTH_CREDIT;
     options.use_authtok = CO_USE_AUTHTOK;
+    options.cluster = CO_CLUSTER;
     memset(options.prompt_type, 0, BUFSIZ);
     strcpy(options.prompt_type,"UNIX");
     options.cracklib_dictpath = CRACKLIB_DICTS;
@@ -642,16 +671,12 @@ PAM_EXTERN int pam_sm_chauthtok(pam_handle_t *pamh, int flags,
 				 options.prompt_type[0]?" ":"");
 
 	    if (retval == PAM_SUCCESS) {     /* a good conversation */
-	        token1 = x_strdup(resp);
+	        token1 = resp;
                 if (token1 == NULL) {
 		    pam_syslog(pamh, LOG_NOTICE,
                                "could not recover authentication token 1");
 		    retval = PAM_AUTHTOK_RECOVERY_ERR;
 		}
-                /*
-                 * tidy up the conversation (resp_retcode) is ignored
-                 */
-                _pam_drop(resp);
             } else {
                 retval = (retval == PAM_SUCCESS) ?
                          PAM_AUTHTOK_RECOVERY_ERR:retval ;
@@ -710,16 +735,12 @@ PAM_EXTERN int pam_sm_chauthtok(pam_handle_t *pamh, int flags,
 				 PROMPT2, options.prompt_type,
 				 options.prompt_type[0]?" ":"");
 	    if (retval == PAM_SUCCESS) {     /* a good conversation */
-	        token2 = x_strdup(resp);
+	        token2 = resp;
 	        if (token2 == NULL) {
 		    pam_syslog(pamh,LOG_NOTICE,
 			       "could not recover authentication token 2");
 		    retval = PAM_AUTHTOK_RECOVERY_ERR;
 		}
-                /*
-                 * tidy up the conversation (resp_retcode) is ignored
-                 */
-	        _pam_drop(resp);
             }
 
 	    /* No else, the a retval == PAM_SUCCESS path can change retval
