@@ -56,6 +56,8 @@
 
 #define PAM_SM_AUTH
 #define PAM_SM_ACCOUNT
+#define PAM_SM_SESSION
+#define PAM_SM_PASSWORD
 
 #include <security/pam_modules.h>
 #include <security/pam_modutil.h>
@@ -184,6 +186,33 @@ evaluate_noglob(const char *left, const char *right)
 {
 	return (fnmatch(right, left, 0) != 0) ? PAM_SUCCESS : PAM_AUTH_ERR;
 }
+/* Check for list match. */
+static int
+evaluate_inlist(const char *left, const char *right)
+{
+	char *p;
+	/* Don't care about left containing ':'. */
+	while ((p=strstr(right, left)) != NULL) {
+		if (p == right || *(p-1) == ':') { /* ':' is a list separator */
+			p += strlen(left);
+			if (*p == '\0' || *p == ':') {
+				return PAM_SUCCESS;
+			}
+		}
+		right = strchr(p, ':');
+		if (right == NULL)
+			break;
+		else
+			++right;
+	}
+	return PAM_AUTH_ERR;
+}
+/* Check for list mismatch. */
+static int
+evaluate_notinlist(const char *left, const char *right)
+{
+	return evaluate_inlist(left, right) != PAM_SUCCESS ? PAM_SUCCESS : PAM_AUTH_ERR;
+}
 /* Return PAM_SUCCESS if the user is in the group. */
 static int
 evaluate_ingroup(pam_handle_t *pamh, const char *user, const char *group)
@@ -250,6 +279,13 @@ evaluate(pam_handle_t *pamh, int debug,
 		snprintf(buf, sizeof(buf), "%s", pwd->pw_dir);
 		left = buf;
 	}
+	if (strcasecmp(left, "service") == 0) {
+		const void *svc;
+		if (pam_get_item(pamh, PAM_SERVICE, &svc) != PAM_SUCCESS)
+			svc = "";
+		snprintf(buf, sizeof(buf), "%s", (const char *)svc);
+		left = buf;
+	}
 	/* If we have no idea what's going on, return an error. */
 	if (left != buf) {
 		pam_syslog(pamh, LOG_CRIT, "unknown attribute \"%s\"", left);
@@ -305,6 +341,13 @@ evaluate(pam_handle_t *pamh, int debug,
 	    (strcasecmp(qual, "noglob") == 0)) {
 		return evaluate_noglob(left, right);
 	}
+	/* Attribute value matches item in list. */
+	if (strcasecmp(qual, "in") == 0) {
+		return evaluate_inlist(left, right);
+	}
+	if (strcasecmp(qual, "notin") == 0) {
+		return evaluate_notinlist(left, right);
+	}
 	/* User is in this group. */
 	if (strcasecmp(qual, "ingroup") == 0) {
 		return evaluate_ingroup(pamh, pwd->pw_name, right);
@@ -376,7 +419,7 @@ pam_sm_authenticate (pam_handle_t *pamh, int flags UNUSED,
 			pam_syslog(pamh, LOG_CRIT,
 				   "error retrieving information about user %lu",
 				   (unsigned long)getuid());
-			return PAM_SERVICE_ERR;
+			return PAM_USER_UNKNOWN;
 		}
 		user = pwd->pw_name;
 	} else {
@@ -395,7 +438,7 @@ pam_sm_authenticate (pam_handle_t *pamh, int flags UNUSED,
 			pam_syslog(pamh, LOG_CRIT,
 				   "error retrieving information about user %s",
 				   user);
-			return PAM_SERVICE_ERR;
+			return PAM_USER_UNKNOWN;
 		}
 	}
 
@@ -468,7 +511,7 @@ PAM_EXTERN int
 pam_sm_setcred(pam_handle_t *pamh UNUSED, int flags UNUSED,
                int argc UNUSED, const char **argv UNUSED)
 {
-	return PAM_SUCCESS;
+	return PAM_IGNORE;
 }
 
 PAM_EXTERN int
@@ -477,6 +520,24 @@ pam_sm_acct_mgmt(pam_handle_t *pamh, int flags, int argc, const char **argv)
 	return pam_sm_authenticate(pamh, flags, argc, argv);
 }
 
+PAM_EXTERN int
+pam_sm_open_session(pam_handle_t *pamh, int flags, int argc, const char **argv)
+{
+	return pam_sm_authenticate(pamh, flags, argc, argv);
+}
+
+PAM_EXTERN int
+pam_sm_close_session(pam_handle_t *pamh, int flags, int argc, const char **argv)
+{
+	return pam_sm_authenticate(pamh, flags, argc, argv);
+}
+
+PAM_EXTERN int
+pam_sm_chauthtok(pam_handle_t *pamh, int flags, int argc, const char **argv)
+{
+	return pam_sm_authenticate(pamh, flags, argc, argv);
+}
+
 /* static module data */
 #ifdef PAM_STATIC
 struct pam_module _pam_succeed_if_modstruct = {
@@ -484,8 +545,8 @@ struct pam_module _pam_succeed_if_modstruct = {
     pam_sm_authenticate,
     pam_sm_setcred,
     pam_sm_acct_mgmt,
-    NULL,
-    NULL,
-    NULL
+    pam_sm_open_session,
+    pam_sm_close_session,
+    pam_sm_chauthtok
 };
 #endif
