@@ -67,10 +67,6 @@ static int selinux_enabled=-1;
 #define SELINUX_ENABLED (selinux_enabled!=-1 ? selinux_enabled : (selinux_enabled=is_selinux_enabled()>0))
 #endif
 
-#ifdef USE_CRACKLIB
-#include <crack.h>
-#endif
-
 #include <security/_pam_macros.h>
 
 /* indicate the following groups are defined */
@@ -86,6 +82,7 @@ static int selinux_enabled=-1;
 #include "support.h"
 #include "passverify.h"
 #include "bigcrypt.h"
+#include "cluster.h"
 
 #if !((__GLIBC__ == 2) && (__GLIBC_MINOR__ >= 1))
 extern int getrpcport(const char *host, unsigned long prognum,
@@ -106,9 +103,6 @@ extern int getrpcport(const char *host, unsigned long prognum,
 #define _UNIX_NEW_AUTHTOK	"-UN*X-NEW-PASS"
 
 #define MAX_PASSWD_TRIES	3
-#ifndef CRACKLIB_DICTS
-#define CRACKLIB_DICTS		NULL
-#endif
 
 static char *getNISserver(pam_handle_t *pamh)
 {
@@ -173,7 +167,7 @@ static int _unix_run_update_binary(pam_handle_t *pamh, unsigned int ctrl, const
         size_t i=0;
         struct rlimit rlim;
 	static char *envp[] = { NULL };
-	char *args[] = { NULL, NULL, NULL, NULL, NULL, NULL };
+	char *args[] = { NULL, NULL, NULL, NULL, NULL, NULL, NULL };
         char buffer[16];
 
 	/* XXX - should really tidy up PAM here too */
@@ -200,10 +194,8 @@ static int _unix_run_update_binary(pam_handle_t *pamh, unsigned int ctrl, const
 	args[0] = x_strdup(UPDATE_HELPER);
 	args[1] = x_strdup(user);
 	args[2] = x_strdup("update");
-	if (on(UNIX_SHADOW, ctrl))
-		args[3] = x_strdup("1");
-	else
-		args[3] = x_strdup("0");
+	args[3] = x_strdup(on(UNIX_SHADOW, ctrl) ? "1" : "0");
+	args[5] = x_strdup(on(UNIX_CLUSTER, ctrl) ? "cluster" : "local");
 
         snprintf(buffer, sizeof(buffer), "%d", remember);
         args[4] = x_strdup(buffer);
@@ -251,7 +243,8 @@ static int _unix_run_update_binary(pam_handle_t *pamh, unsigned int ctrl, const
 }
 #endif
 
-static int check_old_password(const char *forwho, const char *newpass)
+static int check_old_password(unsigned int ctrl,
+			      const char *forwho, const char *newpass)
 {
 	static char buf[16384];
 	char *s_luser, *s_uid, *s_npas, *s_pas;
@@ -306,7 +299,7 @@ static int _do_setpass(pam_handle_t* pamh, const char *forwho,
 		goto done;
 	}
 
-	if (on(UNIX_NIS, ctrl) && _unix_comesfromsource(pamh, forwho, 0, 1)) {
+	if (on(UNIX_NIS, ctrl) && _unix_comesfromsource(pamh, ctrl, forwho, 0, 1)) {
 	    if ((master=getNISserver(pamh)) != NULL) {
 		struct timeval timeout;
 		struct yppasswd yppwd;
@@ -315,7 +308,7 @@ static int _do_setpass(pam_handle_t* pamh, const char *forwho,
 		enum clnt_stat err;
 
 		/* Unlock passwd file to avoid deadlock */
-		unlock_pwdf();
+		unlock_pwdf(ctrl);
 		unlocked = 1;
 
 		/* Initialize password information */
@@ -374,34 +367,34 @@ static int _do_setpass(pam_handle_t* pamh, const char *forwho,
 	    }
 	}
 
-	if (_unix_comesfromsource(pamh, forwho, 1, 0)) {
+	if (_unix_comesfromsource(pamh, ctrl, forwho, 1, 0)) {
 		if(unlocked) {
-			if (lock_pwdf() != PAM_SUCCESS) {
+			if (lock_pwdf(ctrl) != PAM_SUCCESS) {
 				return PAM_AUTHTOK_LOCK_BUSY;
 			}
 		}
 #ifdef WITH_SELINUX
-	        if (unix_selinux_confined())
+	        if (unix_selinux_confined(ctrl))
 			  return _unix_run_update_binary(pamh, ctrl, forwho, fromwhat, towhat, remember);
 #endif
 		/* first, save old password */
-		if (save_old_password(forwho, fromwhat, remember)) {
+		if (save_old_password(ctrl, forwho, fromwhat, remember)) {
 			retval = PAM_AUTHTOK_ERR;
 			goto done;
 		}
 		if (on(UNIX_SHADOW, ctrl) || is_pwd_shadowed(pwd)) {
-			retval = unix_update_shadow(pamh, forwho, towhat);
+			retval = unix_update_shadow(pamh, ctrl, forwho, towhat);
 			if (retval == PAM_SUCCESS)
 				if (!is_pwd_shadowed(pwd))
-					retval = unix_update_passwd(pamh, forwho, "x");
+					retval = unix_update_passwd(pamh, ctrl, forwho, "x");
 		} else {
-			retval = unix_update_passwd(pamh, forwho, towhat);
+			retval = unix_update_passwd(pamh, ctrl, forwho, towhat);
 		}
 	}
 
 
 done:
-	unlock_pwdf();
+	unlock_pwdf(ctrl);
 
 	return retval;
 }
@@ -413,7 +406,7 @@ static int _unix_verify_shadow(pam_handle_t *pamh, const char *user, unsigned in
 	int daysleft;
 	int retval;
 
-	retval = get_account_info(pamh, user, &pwent, &spent);
+	retval = get_account_info(pamh, ctrl, user, &pwent, &spent);
 	if (retval == PAM_USER_UNKNOWN) {
 		return retval;
 	}
@@ -469,16 +462,11 @@ static int _pam_unix_approve_pass(pam_handle_t * pamh
 		}
 	}
 	if (off(UNIX__IAMROOT, ctrl)) {
-#ifdef USE_CRACKLIB
-		remark = FascistCheck (pass_new, CRACKLIB_DICTS);
-		D(("called cracklib [%s]", remark));
-#else
 		if (strlen(pass_new) < 6)
 		  remark = _("You must choose a longer password");
 		D(("length check [%s]", remark));
-#endif
 		if (on(UNIX_REMEMBER_PASSWD, ctrl)) {
-			if ((retval = check_old_password(user, pass_new)) == PAM_AUTHTOK_ERR)
+			if ((retval = check_old_password(ctrl, user, pass_new)) == PAM_AUTHTOK_ERR)
 			  remark = _("Password has been already used. Choose another.");
 			if (retval == PAM_ABORT) {
 				pam_syslog(pamh, LOG_ERR, "can't open %s file to check old passwords",
@@ -545,14 +533,14 @@ PAM_EXTERN int pam_sm_chauthtok(pam_handle_t * pamh, int flags,
 	 * getpwnam() doesn't tell you *where* the information it gives you
 	 * came from, nor should it.  That's our job.
 	 */
-	if (_unix_comesfromsource(pamh, user, 1, on(UNIX_NIS, ctrl)) == 0) {
+	if (_unix_comesfromsource(pamh, ctrl, user, 1, on(UNIX_NIS, ctrl)) == 0) {
 		pam_syslog(pamh, LOG_DEBUG,
-			 "user \"%s\" does not exist in /etc/passwd%s",
-			 user, on(UNIX_NIS, ctrl) ? " or NIS" : "");
+			 "user \"%s\" does not exist in %s%s",
+			 user, PASSWD_FILE, on(UNIX_NIS, ctrl) ? " or NIS" : "");
 		return PAM_USER_UNKNOWN;
 	} else {
 		struct passwd *pwd;
-		_unix_getpwnam(pamh, user, 1, 1, &pwd);
+		_unix_getpwnam(pamh, ctrl, user, 1, 1, &pwd);
 		if (pwd == NULL) {
 			pam_syslog(pamh, LOG_DEBUG,
 				"user \"%s\" has corrupted passwd entry",
@@ -719,7 +707,7 @@ PAM_EXTERN int pam_sm_chauthtok(pam_handle_t * pamh, int flags,
 			pass_new = pass_old = NULL;	/* tidy up */
 			return retval;
 		}
-		if (lock_pwdf() != PAM_SUCCESS) {
+		if (lock_pwdf(ctrl) != PAM_SUCCESS) {
 			return PAM_AUTHTOK_LOCK_BUSY;
 		}
 
@@ -727,7 +715,7 @@ PAM_EXTERN int pam_sm_chauthtok(pam_handle_t * pamh, int flags,
 			retval = _unix_verify_password(pamh, user, pass_old, ctrl);
 			if (retval != PAM_SUCCESS) {
 				pam_syslog(pamh, LOG_NOTICE, "user password changed by another process");
-				unlock_pwdf();
+				unlock_pwdf(ctrl);
 				return retval;
 			}
 		}
@@ -735,7 +723,7 @@ PAM_EXTERN int pam_sm_chauthtok(pam_handle_t * pamh, int flags,
 		retval = _unix_verify_shadow(pamh, user, ctrl);
 		if (retval != PAM_SUCCESS) {
 			pam_syslog(pamh, LOG_NOTICE, "user shadow entry expired");
-			unlock_pwdf();
+			unlock_pwdf(ctrl);
 			return retval;
 		}
 
@@ -744,7 +732,7 @@ PAM_EXTERN int pam_sm_chauthtok(pam_handle_t * pamh, int flags,
 			pam_syslog(pamh, LOG_NOTICE,
 			         "new password not acceptable 2");
 			pass_new = pass_old = NULL;	/* tidy up */
-			unlock_pwdf();
+			unlock_pwdf(ctrl);
 			return retval;
 		}
 
@@ -757,12 +745,12 @@ PAM_EXTERN int pam_sm_chauthtok(pam_handle_t * pamh, int flags,
 		 * First we encrypt the new password.
 		 */
 
-		tpass = create_password_hash(pass_new, ctrl, rounds);
+		tpass = create_password_hash(pamh, pass_new, ctrl, rounds);
 		if (tpass == NULL) {
 			pam_syslog(pamh, LOG_CRIT,
 				"out of memory for password");
 			pass_new = pass_old = NULL;	/* tidy up */
-			unlock_pwdf();
+			unlock_pwdf(ctrl);
 			return PAM_BUF_ERR;
 		}
 
