@@ -28,6 +28,7 @@
 
 #include "support.h"
 #include "passverify.h"
+#include "cluster.c"
 #ifdef WITH_SELINUX
 #include <selinux/selinux.h>
 #define SELINUX_ENABLED is_selinux_enabled()>0
@@ -109,16 +110,8 @@ int _set_ctrl(pam_handle_t *pamh, int flags, int *remember, int *rounds,
 						*remember = 400;
 				}
 			}
-			if (rounds != NULL) {
-				if (j == UNIX_ALGO_ROUNDS) {
-					*rounds = strtol(*argv + 7, NULL, 10);
-					if ((*rounds < 1000) || (*rounds == INT_MAX))
-						/* don't care about bogus values */
-						unset(UNIX_ALGO_ROUNDS, ctrl);
-					if (*rounds >= 10000000)
-						*rounds = 9999999;
-				}
-			}
+			if (rounds != NULL && j == UNIX_ALGO_ROUNDS)
+				*rounds = strtol(*argv + 7, NULL, 10);
 		}
 
 		++argv;		/* step to next argument */
@@ -128,6 +121,31 @@ int _set_ctrl(pam_handle_t *pamh, int flags, int *remember, int *rounds,
 		D(("DISALLOW_NULL_AUTHTOK"));
 		set(UNIX__NONULL, ctrl);
 	}
+	
+	/* Set default rounds for blowfish */
+	if (on(UNIX_BLOWFISH_PASS, ctrl) && off(UNIX_ALGO_ROUNDS, ctrl)) {
+		*rounds = 5;
+		set(UNIX_ALGO_ROUNDS, ctrl);
+	}
+	
+	/* Enforce sane "rounds" values */
+	if (on(UNIX_ALGO_ROUNDS, ctrl)) {
+		if (on(UNIX_BLOWFISH_PASS, ctrl)) {
+			if (*rounds < 4 || *rounds > 31)
+				*rounds = 5;
+		} else if (on(UNIX_SHA256_PASS, ctrl) || on(UNIX_SHA512_PASS, ctrl)) {
+			if ((*rounds < 1000) || (*rounds == INT_MAX))
+				/* don't care about bogus values */
+				unset(UNIX_ALGO_ROUNDS, ctrl);
+			if (*rounds >= 10000000)
+				*rounds = 9999999;
+		}
+	}
+	
+	/* Set unix filename paths */
+	if (on(UNIX_CLUSTER, ctrl))
+		read_cluster_rootdir(pamh);
+	init_unix_files();
 
 	/* auditing is a more sensitive version of debug */
 
@@ -234,7 +252,7 @@ static void _unix_cleanup(pam_handle_t *pamh UNUSED, void *data, int error_statu
 	free(data);
 }
 
-int _unix_getpwnam(pam_handle_t *pamh, const char *name,
+int _unix_getpwnam(pam_handle_t *pamh, unsigned int ctrl, const char *name,
 		   int files, int nis, struct passwd **ret)
 {
 	FILE *passwd;
@@ -246,7 +264,7 @@ int _unix_getpwnam(pam_handle_t *pamh, const char *name,
 
 	if (!matched && files) {
 		int userlen = strlen(name);
-		passwd = fopen("/etc/passwd", "r");
+		passwd = fopen(PASSWD_FILE, "r");
 		if (passwd != NULL) {
 			while (fgets(buf, sizeof(buf), passwd) != NULL) {
 				if ((buf[userlen] == ':') &&
@@ -380,9 +398,9 @@ int _unix_getpwnam(pam_handle_t *pamh, const char *name,
  *
  */
 int _unix_comesfromsource(pam_handle_t *pamh,
-			  const char *name, int files, int nis)
+			  unsigned int ctrl, const char *name, int files, int nis)
 {
-	return _unix_getpwnam(pamh, name, files, nis, NULL);
+	return _unix_getpwnam(pamh, ctrl, name, files, nis, NULL);
 }
 
 /*
@@ -423,7 +441,7 @@ static int _unix_run_helper_binary(pam_handle_t *pamh, const char *passwd,
         int i=0;
         struct rlimit rlim;
 	static char *envp[] = { NULL };
-	char *args[] = { NULL, NULL, NULL, NULL };
+	char *args[] = { NULL, NULL, NULL, NULL, NULL };
 
 	/* XXX - should really tidy up PAM here too */
 
@@ -448,11 +466,9 @@ static int _unix_run_helper_binary(pam_handle_t *pamh, const char *passwd,
 	/* exec binary helper */
 	args[0] = strdup(CHKPWD_HELPER);
 	args[1] = x_strdup(user);
-	if (off(UNIX__NONULL, ctrl)) {	/* this means we've succeeded */
-	  args[2]=strdup("nullok");
-	} else {
-	  args[2]=strdup("nonull");
-	}
+	/* off(UNIX__NONULL, ctrl) means we've succeeded */
+	args[2]=strdup(off(UNIX__NONULL, ctrl) ? "nullok" : "nonull");
+	args[3]=strdup(on(UNIX_CLUSTER, ctrl) ? "cluster" : "local");
 
 	execve(CHKPWD_HELPER, args, envp);
 
@@ -520,7 +536,7 @@ _unix_blankpasswd (pam_handle_t *pamh, unsigned int ctrl, const char *name)
 
 	/* UNIX passwords area */
 
-	retval = get_pwd_hash(pamh, name, &pwd, &salt);
+	retval = get_pwd_hash(pamh, ctrl, name, &pwd, &salt);
 
 	if (retval == PAM_UNIX_RUN_HELPER) {
 		/* salt will not be set here so we can return immediately */
@@ -570,7 +586,7 @@ int _unix_verify_password(pam_handle_t * pamh, const char *name
 
 	D(("locating user's record"));
 
-	retval = get_pwd_hash(pamh, name, &pwd, &salt);
+	retval = get_pwd_hash(pamh, ctrl, name, &pwd, &salt);
 
 	data_name = (char *) malloc(sizeof(FAIL_PREFIX) + strlen(name));
 	if (data_name == NULL) {
@@ -664,17 +680,19 @@ int _unix_verify_password(pam_handle_t * pamh, const char *name
 							    &tty);
 
 					pam_syslog(pamh, LOG_NOTICE,
-					         "authentication failure; "
-					         "logname=%s uid=%d euid=%d "
-					         "tty=%s ruser=%s rhost=%s "
-					         "%s%s",
+					         "authentication failure;"
+					         " logname=%s uid=%d euid=%d"
+					         " tty=%s ruser=%s rhost=%s"
+					         "%s%s"
+					         " pwfile=%s",
 					         new->name, new->uid, new->euid,
 					         tty ? (const char *)tty : "",
 					         ruser ? (const char *)ruser : "",
 					         rhost ? (const char *)rhost : "",
 					         (new->user && new->user[0] != '\0')
 					          ? " user=" : "",
-					         new->user
+					         new->user,
+					         PASSWD_FILE
 					);
 					new->count = 1;
 				}
