@@ -29,15 +29,16 @@
 #include <security/_pam_macros.h>
 
 #include "passverify.h"
+#include "cluster.c"
 
-static int _check_expiry(const char *uname)
+static int _check_expiry(unsigned int ctrl, const char *uname)
 {
 	struct spwd *spent;
 	struct passwd *pwent;
 	int retval;
 	int daysleft;
 
-	retval = get_account_info(uname, &pwent, &spent);
+	retval = get_account_info(ctrl, uname, &pwent, &spent);
 	if (retval != PAM_SUCCESS) {
 		helper_log_err(LOG_ALERT, "could not obtain user info (%s)", uname);
 		printf("-1\n");
@@ -58,8 +59,10 @@ int main(int argc, char *argv[])
 {
 	char pass[MAXPASS + 1];
 	char *option;
-	int npass, nullok;
+	int npass, nullok = 0;
 	int blankpass = 0;
+	int chkexpiry = 0;
+	unsigned int ctrl = 0;
 	int retval = PAM_AUTH_ERR;
 	char *user;
 	char *passwords[] = { pass };
@@ -78,7 +81,7 @@ int main(int argc, char *argv[])
 	 * account).
 	 */
 
-	if (isatty(STDIN_FILENO) || argc != 3 ) {
+	if (isatty(STDIN_FILENO) || argc != 4 ) {
 		helper_log_err(LOG_NOTICE
 		      ,"inappropriate use of Unix helper binary [UID=%d]"
 			 ,getuid());
@@ -101,7 +104,10 @@ int main(int argc, char *argv[])
 	  /* if the caller specifies the username, verify that user
 	     matches it */
 	  if (strcmp(user, argv[1])) {
-	    return PAM_AUTH_ERR;
+	    user = argv[1];
+	    /* no match -> permanently change to the real user and proceed */
+	    if (setuid(getuid()) != 0)
+		return PAM_AUTH_ERR;
 	  }
 	}
 
@@ -109,7 +115,7 @@ int main(int argc, char *argv[])
 
 	if (strcmp(option, "chkexpiry") == 0)
 	  /* Check account information from the shadow file */
-	  return _check_expiry(argv[1]);	  
+	  chkexpiry = 1;
 	/* read the nullok/nonull option */
 	else if (strcmp(option, "nullok") == 0)
 	  nullok = 1;
@@ -117,6 +123,18 @@ int main(int argc, char *argv[])
 	  nullok = 0;
 	else
 	  return PAM_SYSTEM_ERR;
+	
+	if (strcmp(argv[3], "cluster") == 0) {
+		ctrl = UNIX_CLUSTER;
+		if (read_cluster_rootdir() < 0)
+			return PAM_SYSTEM_ERR;
+	} if (strcmp(argv[3], "local") != 0)
+		return PAM_SYSTEM_ERR;
+	
+	init_unix_files();
+	
+	if (chkexpiry)
+		return _check_expiry(ctrl, argv[1]);
 
 	/* read the password from stdin (a pipe from the pam_unix module) */
 
@@ -131,7 +149,7 @@ int main(int argc, char *argv[])
 		blankpass = 1;
 	}
 
-	retval = helper_verify_password(user, pass, nullok);
+	retval = helper_verify_password(ctrl, user, pass, nullok);
 
 	memset(pass, '\0', MAXPASS);	/* clear memory of the password */
 
